#!/usr/bin/env python2.2
########################################
# The contents of this file are subject to the MLX PUBLIC LICENSE version
# 1.0 (the "License"); you may not use this file except in
# compliance with the License.
# 
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
# the License for the specific language governing rights and limitations
# under the License.
# 
# The Original Source Code is "compClust", released 2003 September 03.
# 
# The Original Source Code was developed by the California Institute of
# Technology (Caltech).  Portions created by Caltech are Copyright (C)
# 2002-2003 California Institute of Technology. All Rights Reserved.
########################################

"""
Test suite for the Global Labeling class.
"""

import unittest
import os

from compClust.mlx.datasets import Dataset
from compClust.mlx.views import *
from compClust.mlx.labelings.GlobalLabeling import *

import compClust.mlx
import MLab

class LabelingTestCases(unittest.TestCase):
  def setUp(self):
    """Construct a model to play with"""
    self.original_dir    = os.getcwd()
    os.chdir(compClust.mlx.__path__[0])
    
    data = [ [ 0, 0, 0, 0],
             [ 3, 3, 3, 3],
             [ 1, 1, 1, 1],
             [ 2, 2, 2, 2],
             [ 1, 1, 1, 1],
             [ 3, 3, 3, 3],
             [ 3, 3, 3, 3],
             [ 2, 2, 2, 2],
             [ 2, 2, 2, 2],
             [ 3, 3, 3, 3] ]
    # data = data * 10
    self.dataset = Dataset(data)

  def tearDown(self):
    os.chdir(self.original_dir)

  def testCast(self):

    cv = CachedView(self.dataset)
    l = Labeling(cv)
    l.labelRows([1,2,2,3,4,3,5,5,'a','a'])

    gl = castToGlobalLabeling(l, 'gFoo')
    gl2 = self.dataset.getLabelings()[0]

    assert gl2.getLabelByRows() == l.getLabelByRows()

  def testWithAggregate(self):

    cv = CachedView(self.dataset)
    gl = GlobalWrapper(cv)
    l  = Labeling(cv)

    l.labelRows([1,2,2,3,4,3,5,5,'a','a'])
    gl.labelRows([1,2,2,3,4,3,5,5,'a','a'])

    v1 = RowAggregateFunctionView(cv, gl, MLab.mean)
    v2 = RowAggregateFunctionView(cv, l, MLab.mean)

    assert v1.getData() == v2.getData()
    
  def testLabelUsing(self):

    cv = CachedView(self.dataset)
    gl = GlobalWrapper(self.dataset)

    gl.labelRows(['a','b','c','d','e','f','g','h','i','j'])
    l = Labeling(cv)
    l.labelFrom(gl)
    
    assert gl.getLabelByRows() == l.getLabelByRows()
    
  def simple(self):

    cv1 = CachedView(self.dataset)
    cv2 = CachedView(self.dataset)

    gl = GlobalLabeling(self.dataset)

    gl.addLabelToRow(cv1, 'x', 5)

    assert gl.getRowsByLabel(cv2, 'x') == [5]
    
    self.dataset.removeLabeling(gl)
    self.dataset.removeView(cv1)
    self.dataset.removeView(cv2)

  def medium(self):

    #
    # Test this config:
    #
    #    A   B
    #   / \ / \
    #  C   D   E
    #      
    #      ^
    #      |
    #      +------ create global labeling here
    #

    ds1 = Dataset(MLab.rand(10,5))
    ds2 = Dataset(MLab.rand(10,5))

    ds1.setName("ds1")
    ds2.setName("ds2")

    ss = ColumnSupersetView(ds1, ds2)
    cv1 = CachedView(ds1)
    cv2 = CachedView(ds2)
    sv = SortedView(cv2)
    
    cv1.setName("cv1")
    cv2.setName("cv2")
    sv.setName("sv")
    
    gl = GlobalLabeling(ss, 'foobar')
    gw = GlobalWrapper(ss, glabeling=gl)
    
    gw.addLabelToRow('x', 2)

    assert gl.getRowsByLabel(ds1, 'x') == [2]
    assert gl.getRowsByLabel(ds2, 'x') == [2]
    assert gl.getRowsByLabel(cv1, 'x') == [2]
    assert gl.getRowsByLabel(cv1, 'x') == [2]
    assert gl.getRowsByLabel(ss, 'x')  == [2]

    sv.permuteRows([9,8,7,6,5,4,3,2,1,0])
    assert gl.getRowsByLabel(sv, 'x') == [7]

    gw = cv1.getLabelings()[0]
    gw.addLabelToRow('y', 3)

    assert gl.getRowsByLabel(cv2, 'y') == []
    assert gl.getRowsByLabel(ss, 'y') == [3]
    
    ss.removeLabeling(gl)

    cv2.removeView(sv)
    ds1.removeView(cv1)
    ds2.removeView(cv2)
    ds1.removeView(ss)
    ds2.removeView(ss)

  def medium2(self):

    ds = Dataset(MLab.rand(10,3))

    ss1 = RowSubsetView(ds, [1,2,3,5,8,9])
    ss2 = RowSubsetView(ds, [0,2,4,6,7,8])

    gw = GlobalWrapper(ss1)
    gw.addLabelToRows('x', [0,1,2,3,4,5])

    l = ss2.getLabelings()[0].getRowsByLabel('x')
    l.sort()
    assert l == [1,5]

    ds.removeView(ss1)
    ds.removeView(ss2)
    
  def check_constructor(self):
    labeling = GlobalWrapper(self.dataset)
    self.dataset.removeLabeling(labeling)

  def check_label_retrieval(self):
    labeling = GlobalWrapper(self.dataset)

    labeling.addLabelToKey("foo", 1)
    labeling.addLabelToRow("r", 0)
    labeling.addLabelToCol("c", 0)
    labeling.addLabelToKeys("foo", [3,4])
    labeling.addLabelToRows("r", [5,6])
    labeling.addLabelToCols("c", [1,2])

    labs = labeling.getLabels()

    for lab in labs:
      if lab not in ['foo','r','c']:
        self.fail("Bad label")
    if len(labs) != 3:
      self.fail("Wrong number of labels")

    if labeling.getLabelsByRow(1) != ['foo']:
      self.fail("foo, row 1")
    if labeling.getLabelsByRow(0) != ['r']:
      self.fail("r, row 0")
    if labeling.getLabelsByRow(6) != ['r']:
      self.fail("r, row 6") 
    if labeling.getLabelsByRow(2) != []:
      self.fail("empty, row 2")

    if labeling.getLabelsByCol(0) != ['c']:
      self.fail("c, col 0")
    if labeling.getLabelsByCol(1) != ['c']:
      self.fail("c, col 1")
    if labeling.getLabelsByCol(2) != ['c']:
      self.fail("c, col 2") 
    if labeling.getLabelsByCol(3) != []:
      self.fail("empty, col 3")

    tmp = labeling.getRowsByLabel("r")
    tmp.sort()
    if tmp != [0,5,6]:
      self.fail("label r")
      
    tmp = labeling.getColsByLabel("c")
    tmp.sort()  
    if tmp != [0,1,2]:
      self.fail("label c")

    tmp = labeling.getKeysByLabel("foo")
    tmp.sort()
    if tmp != [1,3,4]:
      self.fail("label foo")

    if labeling.getLabelByRows() != ['r', 'foo', None, 'foo', 'foo', 'r', 'r', None, None, None]:
      self.fail()
      
    self.dataset.removeLabeling(labeling)
    
  def check_name(self):

    rows = range(self.dataset.getNumRows())
    
    v = RowSubsetView(self.dataset, rows)
    l1 = GlobalWrapper(v)
    l1.setName("foo")
    
    l2 = GlobalWrapper(v)
    l2.setName("bar")
    
    l3 = GlobalWrapper(v)
    l3.setName("baz")

    s1 = filter(lambda x : x%4 == 0, rows)
    s2 = filter(lambda x : x%4 == 1, rows)
    s3 = filter(lambda x : x%4 == 2, rows)

    l1.addLabelToRows("1",s1)
    l2.addLabelToRows("2",s2)
    l3.addLabelToRows("3",s3)

    l = v.getLabeling("foo")
    assert l.getLabels()[0] == "1"
    l = v.getLabeling("bar")
    assert l.getLabels()[0] == "2"
    l = v.getLabeling("baz")
    assert l.getLabels()[0] == "3"
    
  def check_sort(self):

    rows = range(self.dataset.getNumRows())

    v = SortedView(self.dataset)
    labeling = GlobalWrapper(v)

    s1 = filter(lambda x : x%4 == 0, rows)
    s2 = filter(lambda x : x%4 == 1, rows)
    s3 = filter(lambda x : x%4 == 2, rows)
    s4 = filter(lambda x : x%4 == 3, rows)

    labeling.addLabelToRows("1",s1)
    labeling.addLabelToRows("2",s2)
    labeling.addLabelToRows("3",s3)
    labeling.addLabelToRows("4",s4)

    labeling.sortDatasetByLabel()
    assert labeling.getLabelByRows() == ['1']*3 + ['2']*3 + ['3']*2 + ['4']*2

  def check_remove_labels(self):
    
    labeling = GlobalWrapper(self.dataset)
    
    labeling.addLabelToRow('a',0)
    labeling.addLabelToRow('b',0)
    labeling.addLabelToRow('c',0)

    labeling.addLabelToCol('a',0)
    labeling.addLabelToCol('b',0)
    labeling.addLabelToCol('c',0)

    assert labeling.getLabelsByRow(0) == ['a','b','c']
    assert labeling.getLabelsByCol(0) == ['a','b','c']

    labeling.removeLabelFromRow('b', 0)
    assert labeling.getLabelsByRow(0) == ['a','c']

    labeling.removeLabel('a')
    assert labeling.getLabelsByRow(0) == ['c']
    assert labeling.getLabelsByCol(0) == ['b','c']

    labeling.removeLabelFromCol('c', 0)
    assert labeling.getLabelsByCol(0) == ['b']

    labeling.removeAll()
    assert labeling.getLabelsByRow(0) == []
    assert labeling.getLabelsByCol(0) == []
    
    self.dataset.removeLabeling(labeling)
    
  def check_create_labels(self):
    labeling = GlobalWrapper(self.dataset)

    labeling.addLabelToKey("foo", 1)
    labeling.addLabelToRow("r", 0)
    labeling.addLabelToCol("c", 0)
    labeling.addLabelToKeys("foo", [3,4])
    labeling.addLabelToRows("r", [5,6])
    labeling.addLabelToCols("c", [1,2])
    
    self.dataset.removeLabeling(labeling)

  def check_addLabel(self):
    labeling = GlobalWrapper(self.dataset)

    keys = self.dataset.getKeys()

    even = filter(lambda x : keys.index(x) & 1 == 0, keys)
    odd  = filter(lambda x : keys.index(x) & 1 == 1, keys)

    labeling.addLabels("foo", even)
    labeling.addLabels("bar", odd)

    assert labeling.getKeysByLabel("foo") == even
    assert labeling.getKeysByLabel("bar") == odd
    
    self.dataset.removeLabeling(labeling)


  def testSuperset(self):

    ds1 = Dataset([[1,2,3],[2,3,4],[3,4,5]])
    ds2 = Dataset([[1,2,3],[2,3,4],[3,4,5]])

    l = GlobalWrapper(ds1, name='names')
    l.labelRows(['d','e','f'])

    l = GlobalWrapper(ds2 , name='names')
    l.labelRows(['a','b','c'])

    v = RowSupersetView(ds1, ds2)

    labels = v.getLabelings()

    assert len(labels) == 2
    assert labels[0].getName() == 'names'
    assert labels[1].getName() == 'names'

def suite():
  suite = unittest.TestSuite()
  suite.addTest(LabelingTestCases("check_constructor"))
  suite.addTest(LabelingTestCases("check_create_labels"))
  suite.addTest(LabelingTestCases("check_label_retrieval"))
  suite.addTest(LabelingTestCases("check_remove_labels"))
  suite.addTest(LabelingTestCases("check_sort"))
  suite.addTest(LabelingTestCases("check_name"))
  suite.addTest(LabelingTestCases("simple"))
  suite.addTest(LabelingTestCases("medium"))
  suite.addTest(LabelingTestCases("medium2"))
  suite.addTest(LabelingTestCases("testLabelUsing"))
  suite.addTest(LabelingTestCases("testWithAggregate"))
  suite.addTest(LabelingTestCases("testCast"))
  suite.addTest(LabelingTestCases("testSuperset"))
  return suite

if __name__ == "__main__":
  unittest.main(defaultTest="suite")
  



