########################################
# The contents of this file are subject to the MLX PUBLIC LICENSE version
# 1.0 (the "License"); you may not use this file except in
# compliance with the License.
# 
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
# the License for the specific language governing rights and limitations
# under the License.
# 
# The Original Source Code is "compClust", released 2003 September 03.
# 
# The Original Source Code was developed by the California Institute of
# Technology (Caltech).  Portions created by Caltech are Copyright (C)
# 2002-2003 California Institute of Technology. All Rights Reserved.
########################################

"""
Utility functions that are useful for accessing files and directories.
"""

import re
import Numeric
import MA

def getFilesByRegExp(dir_name, regexp):
  """list_of_files = getFilesByRegExp(dir_name, rexexp)

  Given a directory and regular expression; find all of the
  files in that directory that match the regular expression.
  """

  import os
  
  files = os.listdir(dir_name)
  filtered_files = filter(lambda x, r = regexp: re.search(r, x), files)

  return filtered_files

###############################################################################
#
# File operations for label files
#
###############################################################################

def readLabelFile(filename, delimiter="\t"):

  import os
  import sys
  import os.path
  import urlparse
  import urllib

  data = None

  if sys.platform == 'win32':
    address_scheme = ''
  else:
    address_scheme = urlparse.urlparse(filename)[0]
  if address_scheme != '':
    (filename, header) = urllib.urlretrieve(filename)

  (root, ext) = os.path.splitext(filename)

  #
  # Extension    Action
  # ---------    ------
  #  .gz         Gzip file
  #  .*          Text file
  
  if ext == ".gz":
    data = readGzipLabelFile(filename, delimiter)
  else:
    data = readTextLabelFile(filename, delimiter)

  if address_scheme != '':
    os.remove(filename)
    
  return data

def readGzipLabelFile(filename, delimiter="\t"):

  import gzip

  stream = gzip.open(filename, 'rb')
  labels = readLabelStream(stream, delimiter)
  stream.close()

  return labels

def readTextLabelFile(filename, delimiter="\t"):

  stream = open(filename)
  labels = readLabelStream(stream, delimiter)
  stream.close()

  return labels

def readLabelStream(stream, delimiter="\t"):

  return __readLabels(stream, delimiter)

def __readLabels(stream, delimiter):

  import string
  
  newline   = '\n'
  empty     = ''
  
  pattern = re.compile('(' + delimiter + '|' + newline + ')')

  invalid = {}

  invalid[delimiter] = 0
  invalid[newline]   = 0
  invalid[empty]     = 0
  
  labels = []  

  while (1):
    
    line = stream.readline()

    if line == '':
      break
  
    strippedLine = string.strip(line)
    

    #
    # If a blank line is encountered, it is the same as no label(s)
    #
      
    if strippedLine == '':
      labels.append([])
      continue

    #
    # Remove the undesirable tokens
    #

    tokens = filter(lambda x : invalid.get(x, 1), pattern.split(line))
    tokens = map(string.strip, tokens)
    labels.append(tokens)
    
  return labels

    
#############################################################################  
#
# File operations for dataset files
#
#############################################################################

def readDelimitedFile(filename, delimiter="\t"):
  
  import os
  import sys
  import os.path
  import urlparse
  import urllib

  data = None
    
  #
  # Check to see if it is a URI address.  If so, copy the file locally
  #

  if sys.platform == 'win32':
    address_scheme = ''
  else:
    address_scheme = urlparse.urlparse(filename)[0]
  if address_scheme != '':
    (filename, header) = urllib.urlretrieve(filename)

  #
  # Attempt to guess the type of file
  #

  (root, ext) = os.path.splitext(filename)

  #
  # Extension    Action
  # ---------    ------
  #  .gz         Gzip file
  #  .*          Text file
  
  if ext == ".gz":
    data = readDelimitedGzipFile(filename, delimiter)
  else:
    data = readDelimitedTextFile(filename, delimiter)

  #
  # Remove the temp file if needed
  #

  if address_scheme != '':
    os.remove(filename)

  return data
  
def readDelimitedGzipFile(filename, delimiter="\t"):

  import gzip

  stream = gzip.open(filename, 'rb')
  data = readDelimitedData(stream, delimiter)
  stream.close()

  return data

def readDelimitedTextFile(filename, delimiter="\t"):

  stream = open(filename)
  data = readDelimitedData(stream, delimiter)
  stream.close()

  return data

def readDelimitedData(stream, delimiter="\t"):
  """
  Attempts to read a dataset which is a text file of floating point values.
  There should be one line per row of data and each datum separated by a
  delimiter.
  """

  return __readData(stream, delimiter)


def __readData(stream, delimiter="\t", flag=0):
  """
  Reads in a dataset and accounts for missing values/NaNs as well.  It will
  assign a masked array to self.data after successfully reading in the data.
  The flag indicates if one should strip the first column (true) or not.
  """

  import string
  
  newline   = '\n'
  empty     = ''
  
  #
  # Enclosing the regular expression in parentheses ensures that
  # delimiters and newlines will be returned along with all other
  # tokens as a result of the split() method.
  #
  
  pattern = re.compile('(' + delimiter + '|' + newline + ')')

  #
  # Make a hash of invalid tokens
  #

  invalid = {}

  invalid[delimiter] = 0
  invalid[newline]   = 0
  invalid[empty]     = 0
  
  #
  # Process each line in stream (filename).
  #
  
  data = []  
  while (1):
    
    line = stream.readline()

    #
    # An empty line means EOF
    #

    if line == '':
      break
  
    strippedLine = string.strip(line)
    
    #
    # Handle blank lines (presumably at the end of a file).
    #
    # If a blank line is encountered, do not attempt to process it.
    # Instead assume the entire dataset has been read.  Break and
    # construct the Numeric array.
    #
      
    if strippedLine == '' or strippedLine[0] == '#':
      continue

    #
    # Remove the undesirable tokens
    #

    tokens = filter(lambda x : invalid.get(x, 1), pattern.split(line))
    
    #
    # Convert to a list of floats and add to the list of datum
    #

    if flag:
      data.append(map(float, tokens[1:]))
    else:

      #
      # Try to be smart.  If the first column is not a number, even if
      # the flag was not set, we'll set the flag ourselves
      #
      
      try:
        data.append(map(float, tokens))
      except ValueError:
        data.append(map(float, tokens[1:]))
        flag = 1
                

  #
  # Construct the Numeric array, a, from the list of lists, data.
  #
  # If the Numeric array is really a vector, maintain the 2-D structure
  # since the dataset schema is based on the data being in a 2-D matrix.
  #
  
  a = Numeric.array(data)
  
  #
  # Construct a mask array, m, containing 1's anywhere where a NaN
  # exists in a, and 0's everywhere else.  According to IEEE-754, NaN
  # is the only floating-point value where NaN != NaN is true.  Use m
  # as the mask parameter when constructing the MaskedArray.
  #
    
  m = Numeric.not_equal(a, a)
  
  if Numeric.sum(Numeric.sum(m)) != 0:
    a = MA.MaskedArray(a, copy=0, mask=m)
    
  return a
  
